""" Code to load the OSIRIS Level 0 data from the standard HDF4 / onyx files.

    At present, readers only exist for the IR level0 data. The others will be added on an as needed basis.

    The methods open_level0_*() all perform the same basic function
    of opening an OSIRIS Level 0 HDF4 data file, reading the data and
    organizing it into an xarray dataset for convenience / usability.

    There are optional parameters to load audit / extra data on some of the methods.

"""
from typing import Dict, List, NamedTuple
import os
from pathlib import Path
import warnings
import xarray as xr
import numpy as np
from numpy.testing import assert_equal, assert_array_equal
from osirisl1services.pyhdfadapter import HDF4VS

# If the attitude file orbit start time differs from the corresponding IR or spectrograph data start time by more than
# this time, load the previous orbit's data as well.
MAX_ATTITUDE_DATA_TIME_DELTA = 5.0 / 24.0 / 3600.0  # 5 seconds (mjd)


class L0Orbit(int):
    """ L0 Orbit class with directory helpers.

        :param int Orbit: the orbit number
    """

    class L0FileParts(NamedTuple):
        prefix: str
        postfix: str

    L0Dict = Dict[str, List[L0FileParts]]

    l0dict: L0Dict = {'IR level0': [L0FileParts(prefix='si0_', postfix='.ir0')],
                      'attitude': [L0FileParts(prefix='so0_', postfix='.oat')],
                      'housekeeping': [L0FileParts(prefix='sh0_', postfix='.hk')],
                      'spectrograph': [L0FileParts(prefix='ss0_', postfix='.os0')]}

    def __new__(cls, value: int):
        if value < 3713:
            warnings.warn('Valid orbits start at 3713.')
        return int.__new__(cls, int(value))

    def as_hex(self):
        return f'{self:04x}'

    def l0_folder(self) -> str:
        """ Return the base folder for the level 0 HDF4 / onyx files. """

        return Path(os.environ['ODINORBITDIR'], '../../Level0/orbit/disk1', f'{self//100*100:05d}')

    def _l0_filename(self, l0type: str) -> str:
        """ Look up the filename for the given Level0 data type. """
        prepost = L0Orbit.l0dict[l0type]
        for (pre, post) in prepost:
            filename = Path(self.l0_folder(), pre + self.as_hex() + post)
            if filename.is_file():
                return filename
        else:
            raise LookupError(f'{l0type} file not found in orbit {self} / {self:x}')

    @property
    def l0_attitude_filename(self):
        return self._l0_filename('attitude')

    @property
    def l0_housekeeping_filename(self):
        return self._l0_filename('housekeeping')

    @property
    def l0_spectrograph_filename(self):
        return self._l0_filename('spectrograph')

    @property
    def l0_ir_filename(self):
        return self._l0_filename('IR level0')


def open_level0_attitude(orbit: int, load_pointing=False, load_ssc_los=False) -> xr.Dataset:
    """ Load the OSIRIS Level 0 attitude data for an orbit. Optinally load the pointing and SSC's LOS data.

        :param Orbit orbit: The orbit class handling the data retrieval, specified by orbit number
        :param int scanno: The scan number to limit the data to, default is None
        :param bool load_pointing: If True, pointing data is included in the dataset
        :param bool load_ssc_los: If True, SSC's line of sight data is included in the dataset

        The returned data structure contains:
            <xarray.Dataset>
            Dimensions:               (Q: 4, ScanNumber: 66, mjd: 5464, xyz: 3)
            Coordinates:
              * mjd                   (mjd) float64 5.239e+04 5.239e+04 5.239e+04 ...
              * xyz                   (xyz) <U1 'x' 'y' 'z'
            Data variables:
                mjdorbitstate         (mjd) float64 5.239e+04 5.239e+04 5.239e+04 ...
                    ** the mjd used to compute the orbit state parameters (position, velocity) **
                stw                   (mjd) int64 598896805 598896822 598896839 ...
                    ** satellite time word **
                position              (mjd, xyz) float64 -4.252e+06 5.54e+06 1.196e+03 ...
                    ** satellite position in ECI coordinate system **
                velocity              (mjd, xyz) float64 823.0 615.0 7.485e+03 828.0 ...
                    ** satellite velocity in ECI coordinate system **
                version               (mjd) float64 17.1 17.1 17.1 17.1 17.1 17.1 17.1 ...
                    ** The current version of satellite software? **

        If the pointing data is loaded, the returned data structure also contains:
            Coordinates:
              * 1ijk           (1ijk) <U1 '1' 'i' 'j' 'k'
            Data variables:
                Qtarget        (mjd, 1ijk) float64 0.7487 0.335 -0.5648 0.0901 0.7488 ...

                Qachieved      (mjd, 1ijk) float64 -0.7487 -0.335 0.5649 -0.0902 -0.7488 ...

                Qerror         (mjd, 1ijk) float64 15.0 5.0 3.0 0.0 15.0 6.0 3.0 0.0 ...


        If the SSC line of sight data is loaded, the returned data structure also contains:
            Data variables:
                osiris_los_latitude   (mjd) float64 23.72 23.78 23.82 23.86 23.91 23.95 ...
                    ** osiris latitude data computed by SSC (differs from level1services) **
                osiris_los_longitude  (mjd) float64 134.9 134.9 134.9 134.9 134.9 134.9 ...
                    ** osiris longitude data computed by SSC (differs from level1 services) **
                osiris_los_height     (mjd) float64 7.989 8.956 9.889 10.76 11.59 12.4 ...
                    ** osiris altitude data computed by SSC (differs from level1 services) **

        For example:

        >>> attitude = open_level0_attitude(orbit=6432)

    """
    with HDF4VS(L0Orbit(orbit).l0_attitude_filename) as vs:
        sa = vs['Spacecraft Attitude']
        ds = xr.Dataset({'mjdorbitstate': (['mjd'], sa['mjdorbitstate']),
                         'stw': (['mjd'], sa['stw']),
                         'position': (['mjd', 'xyz'], sa['position']),
                         'velocity': (['mjd', 'xyz'], sa['velocity']),
                         'version': (['mjd'], sa['version'])},
                        coords={'mjd': sa['mjd'],
                                'xyz': ['x', 'y', 'z']})

        if load_pointing:
            pointing = xr.Dataset({'Qtarget': (['mjd', '1ijk'], sa['Qtarget']),
                                   'Qachieved': (['mjd', '1ijk'], sa['Qachieved']),
                                   'Qerror': (['mjd', '1ijk'], sa['Qerror']),
                                   },
                                  coords={'mjd': sa['mjd'],
                                          '1ijk': ['1', 'i', 'j', 'k']})
            ds = ds.merge(pointing)

        if load_ssc_los:
            line_of_sight = xr.Dataset({'osiris_los_latitude': (['mjd'], sa['osiris_los_latitude']),
                                        'osiris_los_longitude': (['mjd'], sa['osiris_los_longitude']),
                                        'osiris_los_height': (['mjd'], sa['osiris_los_height']),
                                        #  smr data not used by our group
                                        # 'smr_los_latitude': (['mjd'], sa['smr_los_latitude']),
                                        # 'smr_los_longitude': (['mjd'], sa['smr_los_longitude']),
                                        # 'smr_los_height': (['mjd'], sa['smr_los_height']),
                                        },
                                       coords={'mjd': sa['mjd']})
            ds = ds.merge(line_of_sight)
    return ds


def open_level0_ir(orbit: int, channel: int, valid=True, load_audit=False) -> xr.Dataset:
    """ Load the OSIRIS Level 0 IR data for an orbit and channel. Optinally load the audit.

        :param int orbit: The orbit number
        :param int channel: IR channel number to load data from
        :param bool valid: If True, returns the data masked to give only data points without exceptions or flags set
        :param bool load_audit: If True, includes audit data in returned dataset

        The returned data structure contains:
            <xarray.Dataset>
            Dimensions:           (mjd: 2893, pixel: 128, thermistor_store: 3)
            Coordinates:
              * mjd               (mjd) float64 5.239e+04 5.239e+04 ... 5.239e+04 5.239e+04
              * pixel             (pixel) int64 0 1 2 3 4 5 6 ... 122 123 124 125 126 127
            Data variables:
                stw               (mjd) int64 598896817 598896839 ... 598989644 598989676
                exposureTime      (mjd) float64 1.0 1.0 1.0 1.0 1.0 ... 1.0 1.0 1.0 1.0 1.0
                temperature       (mjd) float64 -13.09 -12.96 -12.92 ... -13.09 -13.0 -13.09
                thermistor        (mjd) int64 959 15832 380 ... 15808 380
                mode              (mjd) int64 1294 1294 1294 1294 ... 1294 1294 1294 1294
                scienceprog       (mjd) int64 4092 4092 4092 4092 ... 4092 4092 4092 4092
                shutter           (mjd) int64 0 0 0 0 0 0 0 0 0 0 0 ... 0 0 0 0 0 0 0 0 0 0
                lamp1             (mjd) int64 0 0 0 0 0 0 0 0 0 0 0 ... 0 0 0 0 0 0 0 0 0 0
                lamp2             (mjd) int64 0 0 0 0 0 0 0 0 0 0 0 ... 0 0 0 0 0 0 0 0 0 0
                targetIndex       (mjd) int64 57615 57615 57615 57615 ... 57879 57879 57879
                compressionrate   (mjd) float64 0.8125 0.8125 0.8097 ... 0.6819 0.6797
                exceptions        (mjd) int64 0 0 0 0 0 0 0 0 0 0 0 ... 0 0 0 0 0 0 0 0 0 0
                data              (mjd, pixel) int16 2552 2442 ... 2.934e+03
                orbit             int64 6432
                channel           int64 1
        If the audit data is loaded, the returned data structure also contains:

            Coordinates:
              * store               (store) int64 0 1 2 3

            Data variables:
                audit             (mjd, store) float64 5.369e+04 0.0 0.0 0.0 ... 0.0 0.0 0.0

        For example:

        >>> ir = open_level0_ir(orbit=6432, channel=1)

        Or if you only want the raw data:

        >>> ir_raw = open_level0_ir(orbit=6432, channel=1, valid=False)

    """
    assert 1 <= channel <= 3

    def open_hdf4(o: int) -> xr.Dataset:
        with HDF4VS(L0Orbit(o).l0_ir_filename) as vs:
            l0 = vs[f'IR{channel} Level 0']
            data = np.int16([vs['UI2:UI2:00128']['Array'][i[0]] for i in l0['data']])
            assert_equal([vs['UI2:UI2:00128']['crosscheck'][i[0]] for i in l0['data']], [d[3] for d in l0['data']])
            # error = [vs['R8:R4:00128']['Array'][i[0]] for i in l1['error']]
            # assert_equal([vs['R8:R4:00128']['crosscheck'][i[0]] for i in l1['error']], [d[3] for d in l1['error']])
            # flags = [vs['UI1:UI1:00128']['Array'][i[0]] for i in l1['flags']]
            # assert_equal([vs['UI1:UI1:00128']['crosscheck'][i[0]] for i in l1['flags']], [d[3] for d in l1['flags']])
            ds = xr.Dataset({'stw': (['mjd'], l0['stw']),
                             'exposureTime': (['mjd'], l0['exposureTime']),
                             'temperature': (['mjd'], l0['temperature']),
                             'thermistor': (['mjd'], [i[channel - 1] for i in l0['thermistor']]),
                             'mode': (['mjd'], l0['mode']),
                             'scienceprog': (['mjd'], l0['scienceprog']),
                             'shutter': (['mjd'], l0['shutter']),
                             'lamp1': (['mjd'], l0['lamp1']),
                             'lamp2': (['mjd'], l0['lamp2']),
                             'targetIndex': (['mjd'], l0['targetIndex']),
                             'compressionrate': (['mjd'], l0['compressionrate']),
                             'exceptions': (['mjd'], l0['exceptions']),
                             'data': (['mjd', 'pixel'], data)},
                            coords={'mjd': l0['mjd'],
                                    'pixel': range(128)})

            if load_audit:
                audit = xr.Dataset({'audit': (['mjd', 'store'], l0['audit'])},
                                   coords={'mjd': l0['mjd'],
                                           'store': [0, 1, 2, 3]})
                ds = ds.merge(audit)
            return ds

    ir = open_hdf4(orbit)
    ir['orbit'] = np.int64(orbit)
    ir['channel'] = np.int64(channel)

    # Mask ir to return valid data points if valid is set to true
    if valid:
        ir['data'] = ir.data.where(ir.exceptions == 0)

    return ir
